#ifndef _DEMOLOOPER_H_
#define _DEMOLOOPER_H_

class LoopNode {
public:
	float x,y,z;
	float Wx, Wy, Wz;

    LoopNode(){}
	LoopNode(const float X, const float Y, const float Z, const float WX, const float WY, const float WZ){
		x = X;
		y = Y;
		z = Z;
		Wx = WX;
		Wy = WY;
		Wz = WZ;
	}
};


// ##############################################################################################################################


class DemoLooper {
private:
	CArray <LoopNode *, LoopNode *> nodes;

	/*
		This function performs cubic interpolation between b and c.
		We need a point at both sides to get enough parameters for a cubic curve,
		so a is the point to the left of b, and d is the point to the right of c.
		With x = 0 it returns b, with x = 1 it returns c.
	*/
	static float cerp(const float a, const float b, const float c, const float d, const float x) {
		float p = (d - c) - (a - b);
		float q = (a - b) - p;
		float r = c - a;
		return x * (x * (x * p + q) + r) + b;
	}
	

	/*
		We need this fucntion to allow angles to wrap around, 360 => 0 etc.
		otherwise we'll get a very quick rotation back at the end of the loop
		unless we've rotated equally in both directions.
	*/
	static void fixAngles(float *s){
		if (s[1] > s[0]){
			while (s[1]-s[0] > 180) s[1] -= 360;
		} else {
			while (s[0]-s[1] > 180) s[1] += 360;
		}
		
		if (s[2] > s[1]){
			while (s[2]-s[1] > 180) s[2] -= 360;
		} else {
			while (s[1]-s[2] > 180) s[2] += 360;
		}
		
		if (s[3] > s[2]){
			while (s[3]-s[2] > 180) s[3] -= 360;
		} else {
			while (s[2]-s[3] > 180) s[3] += 360;
		}
	}
public:
	DemoLooper(){}

	~DemoLooper(){
        clear();
	}

    void loadFromFile(char *fileName){
        int i, len;
        LoopNode *node;

        clear();
        FILE *file = fopen(fileName, "rb");
        fread(&len, sizeof(len), 1, file);

        for (i = 0; i < len; i++){
            node = new LoopNode;
            fread(node, sizeof(LoopNode), 1, file);
            nodes.Add(node);
        }
        fclose(file);
    }

    void writeToFile(char *fileName){
        int i, len = nodes.GetSize();

        FILE *file = fopen(fileName, "wb");
        fwrite(&len, sizeof(len), 1, file);
        for (i = 0; i < len; i++){
            fwrite(nodes[i], sizeof(LoopNode), 1, file);
        }
        fclose(file);
    }

	void addNode(const float x, const float y, const float z, const float Wx, const float Wy, const float Wz){
		nodes.Add(new LoopNode(x,y,z,Wx,Wy,Wz));
	}

    void clear(){
		int i, len = nodes.GetSize();
		for (i = 0; i < len; i++){
			delete nodes[i];
		}
        nodes.RemoveAll();
    }

    int getNumberOfNodes(){
        return nodes.GetSize();
    }

	LoopNode getNode(float t){
		int len, tInt;
		float tFrac, ang0, ang1, ang2, s[4];

		tInt = (int) t;
		tFrac = t - tInt;

/*        if (tFrac < 0.5f){
            tFrac = tFrac * tFrac * 2;
        } else {
            tFrac = 1 - (1 - tFrac) * (1 - tFrac);
        }
        */
/*
        d'(0) = 0.5
        d'(1) = 0.5
        d'(0.5) = 1

        d = a * x^3 + b * x^2 + c * x;
        3ax^2 + 2bx + c;

        c = 0.5;
        b = 2;
        a = -4/3;

        3a + 2b = 0;
        3a + 4b = 4;
*/



		len = nodes.GetSize();

		s[0] = nodes[(tInt - 1 + len) % len]->Wx;
		s[1] = nodes[ tInt            % len]->Wx;
		s[2] = nodes[(tInt + 1)       % len]->Wx;
		s[3] = nodes[(tInt + 2)       % len]->Wx;
		fixAngles(s);
		ang0 = cerp(s[0], s[1], s[2], s[3], tFrac);

		s[0] = nodes[(tInt - 1 + len) % len]->Wy;
		s[1] = nodes[ tInt            % len]->Wy;
		s[2] = nodes[(tInt + 1)       % len]->Wy;
		s[3] = nodes[(tInt + 2)       % len]->Wy;
		fixAngles(s);
		ang1 = cerp(s[0], s[1], s[2], s[3], tFrac);

		s[0] = nodes[(tInt - 1 + len) % len]->Wz;
		s[1] = nodes[ tInt            % len]->Wz;
		s[2] = nodes[(tInt + 1)       % len]->Wz;
		s[3] = nodes[(tInt + 2)       % len]->Wz;
		fixAngles(s);
		ang2 = cerp(s[0], s[1], s[2], s[3], tFrac);

		return LoopNode(
			cerp(nodes[(tInt - 1 + len) % len]->x, nodes[tInt % len]->x, nodes[(tInt + 1) % len]->x, nodes[(tInt + 2) % len]->x, tFrac), 
			cerp(nodes[(tInt - 1 + len) % len]->y, nodes[tInt % len]->y, nodes[(tInt + 1) % len]->y, nodes[(tInt + 2) % len]->y, tFrac), 
			cerp(nodes[(tInt - 1 + len) % len]->z, nodes[tInt % len]->z, nodes[(tInt + 1) % len]->z, nodes[(tInt + 2) % len]->z, tFrac), 
			ang0, ang1, ang2);
			
	}

};



#endif // _DEMOLOOPER_H_